from django.contrib.auth import authenticate
from paramiko import ServerInterface, AUTH_SUCCESSFUL, OPEN_SUCCEEDED, AUTH_FAILED
from . import models
from .compat import get_username_field


class StubServer(ServerInterface):
    model = models.SFTPUserAccount
    username_field = get_username_field()
    username = None

    def check_auth_password(self, username, password):
        user = authenticate(
            **{self.username_field: username, 'password': password}
        )
        account = self.get_account(username)
        if not (user and account):
            return AUTH_FAILED
        self.username = username
        return AUTH_SUCCESSFUL

    def check_auth_publickey(self, username, key):
        return AUTH_FAILED

    def check_channel_request(self, kind, chanid):
        return OPEN_SUCCEEDED

    def get_allowed_auths(self, username):
        """List availble auth mechanisms."""
        return "password"

    def _filter_user_by(self, username):
        return {"user__%s" % self.username_field: username}

    def get_account(self, username):
        """return user by username.
        """
        try:
            account = self.model.objects.get(
                **self._filter_user_by(username)
            )
        except self.model.DoesNotExist:
            return None
        return account

    def get_home_dir(self, username=None):
        if not username:
            username = self.username
        account = self.get_account(username)
        if not account:
            return ''
        return account.get_home_dir()




from django.contrib import admin

from . import models


class SFTPUserGroupAdmin(admin.ModelAdmin):
    """Admin class for SFTPUserGroup
    """
    list_display = ('name',)
    search_fields = ('name',)


class SFTPUserAccountAdmin(admin.ModelAdmin):
    """Admin class for SFTPUserAccountAdmin
    """
    list_display = ('user', 'group', 'last_login')
    search_fields = ('user', 'group', 'last_login')


admin.site.register(models.SFTPUserGroup, SFTPUserGroupAdmin)
admin.site.register(models.SFTPUserAccount, SFTPUserAccountAdmin)



# -*- coding: utf-8 -*-
import os

from django.db import models
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from django.utils.encoding import python_2_unicode_compatible


@python_2_unicode_compatible
class SFTPUserGroup(models.Model):
    name = models.CharField(
        _("Group name"), max_length=30, null=False, blank=False, unique=True)
    home_dir = models.CharField(
        _("Home directory"), max_length=1024, null=True, blank=True)

    def __str__(self):
        return u"{0}".format(self.name)

    class Meta:
        verbose_name = _("SFTP user group")
        verbose_name_plural = _("SFTP user groups")


@python_2_unicode_compatible
class SFTPUserAccount(models.Model):
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL, verbose_name=_("User"),
        on_delete=models.CASCADE)
    group = models.ForeignKey(
        SFTPUserGroup, verbose_name=_("SFTP user group"), null=False,
        blank=False, on_delete=models.CASCADE)
    last_login = models.DateTimeField(
        _("Last login"), editable=False, null=True)
    home_dir = models.CharField(
        _("Home directory"), max_length=1024, null=True, blank=True)

    def __str__(self):
        try:
            user = self.user
        except ObjectDoesNotExist:
            user = None
        return u"{0}".format(user)

    def get_username(self):
        try:
            user = self.user
        except ObjectDoesNotExist:
            user = None
        return user and user.username or ""

    def update_last_login(self, value=None):
        self.last_login = value or timezone.now()

    def get_home_dir(self):
        if self.home_dir:
            directory = self.home_dir
        elif self.group and self.group.home_dir:
            directory = self.group.home_dir
        else:
            directory = os.path.join(
                os.path.dirname(os.path.expanduser('~')), u'{username}')
        return directory.format(username=self.get_username())

    def has_perm(self, perm, path):
        return perm in self.get_perms()

    def get_perms(self):
        return self.group.permission

    class Meta:
        verbose_name = _("SFTP user account")
        verbose_name_plural = _("SFTP user accounts")



import os

from paramiko import SFTPServer, SFTPAttributes, \
    SFTPHandle, SFTP_OK


class StubSFTPHandle(SFTPHandle):
    def stat(self):
        try:
            return SFTPAttributes.from_stat(os.fstat(self.readfile.fileno()))
        except OSError as e:
            return SFTPServer.convert_errno(e.errno)

    def chattr(self, attr):
        # python doesn't have equivalents to fchown or fchmod, so we have to
        # use the stored filename
        return SFTP_OK



# -*- coding: utf-8 -*-
# Generated by Django 1.11.19 on 2019-02-11 12:18
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='SFTPUserAccount',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_login', models.DateTimeField(editable=False, null=True, verbose_name='Last login')),
                ('home_dir', models.CharField(blank=True, max_length=1024, null=True, verbose_name='Home directory')),
            ],
            options={
                'verbose_name': 'SFTP user account',
                'verbose_name_plural': 'SFTP user accounts',
            },
        ),
        migrations.CreateModel(
            name='SFTPUserGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=30, unique=True, verbose_name='Group name')),
                ('home_dir', models.CharField(blank=True, max_length=1024, null=True, verbose_name='Home directory')),
            ],
            options={
                'verbose_name': 'SFTP user group',
                'verbose_name_plural': 'SFTP user groups',
            },
        ),
        migrations.AddField(
            model_name='sftpuseraccount',
            name='group',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_sftp.SFTPUserGroup', verbose_name='SFTP user group'),
        ),
        migrations.AddField(
            model_name='sftpuseraccount',
            name='user',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='User'),
        ),
    ]



import os
import time
from collections import namedtuple

from django.core.files.storage import (
    get_storage_class as _get_storage_class
)
from paramiko import SFTPServerInterface, SFTPServer, SFTPAttributes
from paramiko.common import DEBUG
from paramiko.sftp import Message, SFTP_OK


from django_sftp.handler import StubSFTPHandle

PseudoStat = namedtuple(
    'PseudoStat',
    [
        'st_size', 'st_mtime', 'st_nlink', 'st_mode', 'st_uid', 'st_gid',
        'st_dev', 'st_ino', 'st_atime'
    ])


class StubSFTPServer(SFTPServer):

    def start_subsystem(self, name, transport, channel):
        self.sock = channel
        self._log(DEBUG, 'Started sftp server on channel %s' % repr(channel))
        self._send_server_version()
        self.server.session_started()
        while True:
            try:
                t, data = self._read_packet()
            except EOFError:
                return
            except Exception:
                return
            msg = Message(data)
            request_number = msg.get_int()
            try:
                self._process(t, request_number, msg)
            except Exception:
                # send some kind of failure message, at least
                try:
                    self._send_status(request_number, SFTP_OK)
                except:
                    pass


class StubSFTPServerInterface(SFTPServerInterface):
    # assume current folder is a fine root
    # (the tests always create and eventualy delete a subfolder, so there shouldn't be any mess)
    ROOT = '/'  # os.getcwd()
    storage_class = None
    server_interface = None

    def __init__(self, server, *largs, **kwargs):
        """
        Create a new SFTPServerInterface object.  This method does nothing by
        default and is meant to be overridden by subclasses.

        :param .ServerInterface server:
            the server object associated with this channel and SFTP subsystem
        """
        super(StubSFTPServerInterface, self).__init__(server, *largs, **kwargs)
        self.storage = self.get_storage()
        self.server_interface = server
        self.ROOT = self.server_interface.get_home_dir()

    def session_started(self):
        """
        The SFTP server session has just started.  This method is meant to be
        overridden to perform any necessary setup before handling callbacks
        from SFTP operations.
        """
        pass

    def session_ended(self):
        """
        The SFTP server session has just ended, either cleanly or via an
        exception.  This method is meant to be overridden to perform any
        necessary cleanup before this `.SFTPServerInterface` object is
        destroyed.
        """
        pass

    def get_storage_class(self):
        if self.storage_class is None:
            return _get_storage_class()
        return self.storage_class

    def get_storage(self):
        storage_class = self.get_storage_class()
        return storage_class()

    def _realpath(self, path):
        root = self.ROOT
        if self.ROOT.endswith('/'):
            root = self.ROOT[:-1]
        return root + self.canonicalize(path)

    def list_folder(self, path):
        path = self._realpath(path)
        out = []
        directories, files = self.storage.listdir(path)
        flist = [name + '/' for name in directories if name] + [name for name in files if name]
        for fname in flist:
            attr = SFTPAttributes.from_stat(self.stat(os.path.join(path, fname)))
            attr.filename = fname.replace('/', '')
            out.append(attr)
        return out

    def stat(self, path):
        if self.isfile(path):
            st_mode = 0o0100770
        else:
            # directory
            st_mode = 0o0040770
        st_time = int(self.getmtime(path))
        return PseudoStat(
            st_size=self.getsize(path),
            st_mtime=st_time,
            st_atime=st_time,
            st_nlink=1,
            st_mode=st_mode,
            st_uid=1000,
            st_gid=1000,
            st_dev=0,
            st_ino=0,
        )

    lstat = stat

    def _exists(self, path):
        if path.endswith('/'):
            return True
        return self.storage.exists(path)

    def isfile(self, path):

        return self._exists(path) and not path.endswith('/')

    def islink(self, path):
        return False

    def isdir(self, path):
        return not self.isfile(path)

    def getmtime(self, path):
        if self.isdir(path):
            return 0
        return self._origin_getmtime(path)

    def getsize(self, path):
        if self.isdir(path):
            return 0
        return self.storage.size(path)

    def _origin_getmtime(self, path):
        t = time.mktime(self.storage.get_modified_time(path).timetuple())
        return t

    def open(self, path, flags, attr):
        path = self._realpath(path)
        if flags:
            f = self.storage.open(path, 'wb')
            self.server_interface.add_log('uploaded: {}'.format(path))
        else:
            f = self.storage.open(path, 'rb')
            self.server_interface.add_log('downloaded: {}'.format(path))
        print(flags)
        fobj = StubSFTPHandle(flags)
        fobj.filename = path
        fobj.readfile = f
        fobj.writefile = f
        return fobj

    def remove(self, path):
        path = self._realpath(path)
        self.server_interface.add_log('removed: {}'.format(path))
        try:
            self.storage.delete(path)
        except OSError as e:
            return SFTPServer.convert_errno(e.errno)
        return SFTP_OK

    def rename(self, oldpath, newpath):
        # oldpath = self._realpath(oldpath)
        # newpath = self._realpath(newpath)
        # try:
        #     os.rename(oldpath, newpath)
        # except OSError as e:
        #     return SFTPServer.convert_errno(e.errno)
        return SFTP_OK

    def readlink(self, path):
        return SFTP_OK

    def symlink(self, target_path, path):
        return SFTP_OK

    def mkdir(self, path, attr):
        return SFTP_OK

    def rmdir(self, path):
        return SFTP_OK

    def chattr(self, path, attr):
        return SFTP_OK



import logging
import socket
import threading
import time

import paramiko
from django.core.management.base import BaseCommand

from django_sftp.interface import StubServer
from django_sftp.server import StubSFTPServerInterface, StubSFTPServer

BACKLOG = 10

logging.disable(logging.DEBUG)


class ConnHandlerThd(threading.Thread):
    def __init__(self, conn, keyfile, *args, **kwargs):
        super(ConnHandlerThd, self).__init__(*args, **kwargs)
        self._conn = conn
        self._keyfile = keyfile

    def run(self):
        host_key = paramiko.RSAKey.from_private_key_file(self._keyfile)
        transport = paramiko.Transport(self._conn)
        transport.add_server_key(host_key)
        transport.set_subsystem_handler(
            'sftp', StubSFTPServer, StubSFTPServerInterface)

        server = StubServer()
        transport.start_server(server=server)

        channel = transport.accept()
        while transport.is_active():
            time.sleep(1)


def start_server(host, port, keyfile, level):
    paramiko_level = getattr(paramiko.common, level)
    paramiko.common.logging.basicConfig(level=paramiko_level)
    paramiko.common.logging.disable(logging.CRITICAL)

    server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)
    server_socket.bind((host, port))
    server_socket.listen(BACKLOG)

    while True:
        conn, addr = server_socket.accept()
        srv_thd = ConnHandlerThd(conn, keyfile)
        srv_thd.setDaemon(True)
        srv_thd.start()


class Command(BaseCommand):
    help = "Start SFTP server"

    def add_arguments(self, parser):
        parser.add_argument('host_port', nargs="?")
        parser.add_argument(
            '-l', '--level', dest='level', default='DEBUG',
            help='Debug level: WARNING, INFO, DEBUG [default: %(default)s]'
        )
        parser.add_argument(
            '-k', '--keyfile', dest='keyfile', metavar='FILE',
            help='Path to private key, for example /tmp/test_rsa.key'
        )

    def handle(self, *args, **options):
        # bind host and port
        host_port = options.get('host_port')
        host, _port = host_port.split(':', 1)
        port = int(_port)

        level = options['level']
        keyfile = options['keyfile']
        start_server(host, port, keyfile, level)



"""
The `compat` module provides support for backwards compatibility
with older versions of django
"""

from django.conf import settings


def get_user_model_path():
    return getattr(settings, 'AUTH_USER_MODEL', None) or 'auth.User'


def get_username_field():
    from django.contrib.auth import get_user_model
    UserModel = get_user_model()
    return getattr(UserModel, 'USERNAME_FIELD', 'username')
